﻿using System;
using System.IO;
using System.Net;
using System.Net.Sockets;
using System.Runtime.InteropServices;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace ConnectAdventurer3 {
	public class Client {
		/// <summary>
		/// Adventurer3への接続ポート番号(一応固定)
		/// </summary>
		const int Adventurer3Port = 8899;
		StreamWriter output_;
		string ip_;
		System.Net.Sockets.TcpClient tcp_;

		public Client(IPAddress ip, System.IO.StreamWriter output) {
			output_ = output;
			ip_ = ip.ToString();
		}

		private void read(NetworkStream ns) {
			System.Text.Encoding enc = System.Text.Encoding.UTF8;
			// サーバーから送られたデータを受信する
			var resBytes = new byte[256];
			int resSize = 0;
			do {
				resSize = ns.Read(resBytes, 0, resBytes.Length);
				if (resSize == 0) {
					break;
				}
				var work = enc.GetChars(resBytes);
				output_.WriteLine(work);
				output_.Flush();
			} while (ns.DataAvailable);
		}

		public void start() {
			if (tcp_ == null) {
				tcp_ = new TcpClient(ip_, Adventurer3Port);
				var ns = tcp_.GetStream();
				//タイムアウト設定
				ns.ReadTimeout = 10000;
				ns.WriteTimeout = 10000;

				System.Text.Encoding enc = System.Text.Encoding.UTF8;
				byte[] sendBytes = enc.GetBytes("~M601 S1\r\n");
				//データを送信する
				ns.Write(sendBytes, 0, sendBytes.Length);
				read(ns);
			}
		}

		public void send(string cmd) {
			if (tcp_ != null) {
				var ns = tcp_.GetStream();
				System.Text.Encoding enc = System.Text.Encoding.UTF8;
				byte[] sendBytes = enc.GetBytes(string.Format("~{0}\r\n", cmd));
				//データを送信する
				ns.Write(sendBytes, 0, sendBytes.Length);
				read(ns);
			}
		}

		public void stop() {
			if (tcp_ != null) {
				var ns = tcp_.GetStream();
				System.Text.Encoding enc = System.Text.Encoding.UTF8;
				byte[] sendBytes = enc.GetBytes("~M602\r\n");
				//データを送信する
				ns.Write(sendBytes, 0, sendBytes.Length);
				read(ns);
				tcp_.Close();
				tcp_ = null;
			}
		}
	}

	class Program {
		[DllImport("iphlpapi.dll")]
		extern static int GetIpNetTable(IntPtr pTcpTable, ref int pdwSize, bool bOrder);

		[StructLayout(LayoutKind.Sequential)]
		public struct MIB_IPNETROW {
			public int Index;
			public int PhysAddrLen;
			[MarshalAs(UnmanagedType.ByValArray, SizeConst = 8)]
			public byte[] PhysAddr;
			public int Addr;
			public int Type;
		}

		/// <summary>
		/// Adventurer3のIPをサーチするクラス
		/// ARPテーブルから登録済みIPを検索し、そのIPアドレスのMACアドレス先頭02で始まるものを返す
		/// ARPテーブル検索は、GetIpNetTableを使用する
		/// </summary>
		static List<IPAddress> searchAdventurer3IP() {
			var ips = new List<IPAddress>();

			int size = 0;
			GetIpNetTable(IntPtr.Zero, ref size, true);//必要サイズの取得
			var p = Marshal.AllocHGlobal(size);//メモリ割当て
			if (GetIpNetTable(p, ref size, true) == 0) {//データの取得
				var num = Marshal.ReadInt32(p);//MIB_IPNETTABLE.dwNumEntries(データ数)
				var ptr = IntPtr.Add(p, 4);
				for (int i = 0; i < num; i++) {
					MIB_IPNETROW work = (MIB_IPNETROW)Marshal.PtrToStructure(ptr, typeof(MIB_IPNETROW));
					if (work.PhysAddrLen > 0 && work.PhysAddr[0] == 0x02 && work.Type == 3) {
						ips.Add(new IPAddress(BitConverter.GetBytes(work.Addr)));
					}
					ptr = IntPtr.Add(ptr, Marshal.SizeOf(typeof(MIB_IPNETROW)));//次のデータ
				}
				Marshal.FreeHGlobal(p);  //メモリ開放
			}

			return ips;
		}

		static void Main(string[] args) {
			var work = searchAdventurer3IP();
			foreach (var add in work) {
				var stream = System.Console.OpenStandardOutput();
				var client = new Client(add, new StreamWriter(stream));
				client.start();
				client.send("M105");
				client.send("M119");
				client.send("M27");
				client.stop();
			}
		}
	}
}
